import asyncio
import logging
import signal
import sys

from .links import AsyncIPyCLink


class AsyncIPyCHost:
    """Represents an abstracted async socket listener that connects with
    and listens to :class:`AsyncIPyCClient` clients.
    A number of options can be passed to the :class:`AsyncIPyCHost`.

    Parameters
    -----------
    ip_address: Optional[:class:`str`]
        The IP address start listening on. This defaults to ``localhost``.
    port: Optional[:class:`int`]
        The port the listener binds to. This defaults to ``9999``. Check
        your system to make sure this port is not used by another service.
        To use multiple :class:`AsyncIPyCHost` hosts, ensure the ports are
        different between instantiations.
    loop: Optional[:class:`asyncio.AbstractEventLoop`]
        The :class:`asyncio.AbstractEventLoop` to use for asynchronous operations.
        Defaults to ``None``, in which case the default event loop is used via
        :func:`asyncio.get_event_loop()`.

    Attributes
    -----------
    loop: :class:`asyncio.AbstractEventLoop`
        The event loop that the client uses for asynchronous events.
    """
    def __init__(self, ip_address: str='localhost', port:  int=9999, loop=None):
        self._ip_address = ip_address
        self._port = port
        self._logger = logging.getLogger(self.__class__.__name__)
        self.loop = asyncio.get_event_loop() if loop is None else loop
        self._server = None
        self._closed = False
        self._connections = set()
        self._handlers = {
            'connect': set()
        }
        self._on_close = asyncio.Event()

    async def __handle_connection(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        new_connection = AsyncIPyCLink(reader, writer, self)
        self.connections.add(new_connection)
        for handle in self._handlers['connect']:
            await handle(new_connection)

    def _cleanup_loop(self, loop):
        try:
            self._cancel_tasks(loop)
            if sys.version_info >= (3, 6):
                loop.run_until_complete(loop.shutdown_asyncgens())
        finally:
            self._logger.debug('Closing the event loop.')
            loop.close()

    def _cancel_tasks(self, loop):
        try:
            task_retriever = asyncio.Task.all_tasks
        except AttributeError:
            # future proofing for 3.9
            task_retriever = asyncio.all_tasks

        tasks = {t for t in task_retriever(loop=loop) if not t.done()}

        if not tasks:
            return

        self._logger.debug('Cleaning up after %d tasks.', len(tasks))
        for task in tasks:
            task.cancel()

        loop.run_until_complete(asyncio.gather(*tasks, return_exceptions=True))
        self._logger.debug('All tasks finished cancelling.')

        for task in tasks:
            if task.cancelled():
                continue
            if task.exception() is not None:
                loop.call_exception_handler({
                    'message': 'Unhandled exception during AsyncIPyCHost.run shutdown.',
                    'exception': task.exception(),
                    'task': task
                })

    def on_connect(self, coro):
        """A decorator that registers a coroutine to execute when a connection is made to the listener.

        The decorated function must be a :ref:`coroutine <coroutine>` and possess one parameter for the :class:`AsyncIPyCLink` connection link that is supplied on connection; if not, a :exc:`TypeError` is raised.

        Parameters
        ------------
        coro: :ref:`coroutine <coroutine>`
            The coroutine handler to be called when a new connection is made to the listener
        Example
        ---------
        .. code-block:: python3

            host = AsyncIPyCHost()

            @host.on_connect
            async def connectionMade(link: AsyncIPyCLink):
                print('A connection was made!')
        Raises
        --------
        TypeError
            The coroutine passed is not actually a coroutine or does not contain enough arguments.
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('@on_connect must register a coroutine function')

        if coro.__code__.co_argcount not in [1, 2]:
            raise TypeError('@on_connect coroutines must allow for a Link argument')

        self._handlers['connect'].add(coro)
        self._logger.debug(f'[IPyCHost] {coro.__name__} has successfully been registered as an on_connect event')
        return coro

    def add_connection_handler(self, coro):
        """Wrapped decorator for the :meth:`on_connect` method.

        Parameters
        ------------
        coro: :ref:`coroutine <coroutine>`
            The coroutine handler to be called when a new connection is made to the listener
        Raises
        --------
        TypeError
            The coroutine passed is not actually a coroutine or does not contain enough arguments.
        """
        self.on_connect(coro)

    def remove_connection_handler(self, coro):
        """Removes a connection handler from the internal listener dispatcher.

        Parameters
        ------------
        coro: :ref:`coroutine <coroutine>`
            The coroutine handler to be removed.
        """
        if coro in self._handlers['connect']:
            self._handlers['connect'].remove(coro)

    def is_closed(self):
        """:class:`bool`: Indicates if the underlying socket listener is closed or no longer listening."""
        return self._closed

    @property
    def connections(self):
        """:class:`set`: Returns the set of all active :class:`AsyncIPyCLink` connections the host is handling."""
        return self._connections

    async def start(self, *args):
        """|coro|

        A shorthand coroutine for :func:`asyncio.start_server`. Any
        arguments supplied are passed to :func:`asyncio.start_server()`
        and afterward to :func:`asyncio.loop.create_server()`. See the asyncio
        documentation for these arguments and their use.

        """
        self._server = await asyncio.start_server(self.__handle_connection, host=self._ip_address, port=self._port, loop=self.loop, *args)

    def run(self, *args):
        """A blocking call that begins server listening and abstracts
        away the asyncio event loop initialisation and handling.

        If you want more control over the event loop then this
        function should not be used. Use the :meth:`start`
        coroutine instead.

        Any arguments supplied is directly passed to the
        :meth:`start` coroutine. See it's documentation for use.

        .. warning::

            This function must be the last function to call due to the fact that it
            is blocking. That means that anything being called after this function
            or executed will not happen until this host closes or terminates.
        """
        try:
            self.loop.add_signal_handler(signal.SIGINT, lambda: self.loop.stop())
            self.loop.add_signal_handler(signal.SIGTERM, lambda: self.loop.stop())
        except NotImplementedError:
            pass

        async def runner():
            try:
                await self.start(*args)
                await self._on_close.wait()
            finally:
                if not self.is_closed():
                    await self.close()

        def stop_loop_on_completion(_):
            self.loop.stop()

        future = asyncio.ensure_future(runner(), loop=self.loop)
        future.add_done_callback(stop_loop_on_completion)
        try:
            self.loop.run_forever()
        except KeyboardInterrupt:
            self._logger.debug('Received signal to terminate bot and event loop.')
        finally:
            future.remove_done_callback(stop_loop_on_completion)
            self._logger.debug('Cleaning up tasks.')
            self._cleanup_loop(self.loop)

        if not future.cancelled():
            try:
                return future.result()
            except KeyboardInterrupt:
                return None

    async def close(self):
        """|coro|

        Closes all :class:`AsyncIPyCLink` connections and stops the internal listener.
        """
        if self._closed:
            return
        self._closed = True
        for connection in list(self.connections):
            await connection.close()
        self._on_close.set()


class AsyncIPyCMaster(AsyncIPyCHost):
    """Pseudo-class for AsyncIPyCHost"""
    pass


class AsyncIPyCClient:
    """Represents an abstracted async socket client that connects to
    and communicates with :class:`AsyncIPyCHost` hosts.
    A number of options can be passed to the :class:`AsyncIPyCClient`.

    Parameters
    -----------
    ip_address: Optional[:class:`str`]
        The IP address to connect to. This defaults to ``localhost``.
    port: Optional[:class:`int`]
        The port to target at the host IP address. This defaults to ``9999``.
    loop: Optional[:class:`asyncio.AbstractEventLoop`]
        The :class:`asyncio.AbstractEventLoop` to use for asynchronous operations.
        Defaults to ``None``, in which case the default event loop is used via
        :func:`asyncio.get_event_loop()`.

    Attributes
    -----------
    loop: :class:`asyncio.AbstractEventLoop`
        The event loop that the client uses for asynchronous events.
    """
    def __init__(self, ip_address: str='localhost', port:  int=9999, loop=None):
        self._ip_address = ip_address
        self._port = port
        self._logger = logging.getLogger(self.__class__.__name__)
        self.loop = asyncio.get_event_loop() if loop is None else loop
        self._link = None
        self._closed = False
        self._handlers = {
            'connect': []
        }

    async def connect(self, *args) -> AsyncIPyCLink:
        """|coro|

        A shorthand coroutine for :func:`asyncio.open_connection`. Any
        arguments supplied are directly passed to this method; See the
        asyncio documentation for these arguments and their use.

        Returns
        -------
        :class:`AsyncIPyCLink`
            The connection that has been established with a :class:`AsyncIPyCHost`.
        """
        reader, writer = await asyncio.open_connection(host=self._ip_address, port=self._port, loop=self.loop, *args)
        self._link = AsyncIPyCLink(reader, writer, self)
        return self._link

    async def close(self):
        """|coro|

        Closes the :class:`AsyncIPyCLink` connection and stops the client connection.
        """
        if self._closed:
            return
        self._closed = True
        await self._link.close()
        self._link = None

    @property
    def connections(self):
        """:class:`set`: Returns the set of all active :class:`AsyncIPyCLink` connections the client is handling.
        The number of connections is always either one or none.
        """
        return {self._link} if self._link else None


class AsyncIPyCSlave(AsyncIPyCClient):
    """Pseudo-class for AsyncIPyCClient"""
    pass



import asyncio
import logging
import sys

from multiprocessing.connection import Connection

from .packets import CommunicationPacket
from . import serialization


class IPyCLink:
    """Represents an abstracted synchronous socket connection that handles
    communication between a :class:`IPyCHost` and a :class:`IPyCClient`
    This class is internally managed and typically should not be instantiated on
    its own.

    Parameters
    -----------
    connection: :class:`multiprocessing.connection.Connection`
        The managed socket connection.
    client: Union[:class:`IPyCHost`, :class:`IPyCClient`]
        The communication object that is responsible for managing this connection.
    """
    def __init__(self, connection: Connection, client):
        self._connection = connection
        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.debug(f"Established link")
        self._active = True
        self._client = client

    def close(self):
        """Closes the socket channel with a peer and attempts to send them EOF.
        Informs the parent :class:`IPyCHost` or :class:`IPyCClient` of the
        closed connection.
        """
        self._logger.debug(f"Beginning to close link")
        self._connection.close()
        self._active = False
        self._client.connections.remove(self)
        self._logger.debug(f"Closed link")

    def is_active(self):
        """:class:`bool`: Indicates if the socket connection is closed, at EOF, or no longer viable."""
        # Quickly check if the state of the reader changed from the remote
        if not self._connection or self._connection.closed:
            self._active = False
        return self._active

    def send(self, serializable_object: object, encoding='utf-8'):
        """Send a serializable object to the receiving end. If the object is not a custom
        serializable object, python's builtins will be used. If the object is a custom
        serializable, the receiving end must also have this object in their list of custom
        deserializers.

        .. warning::
            After the result of serialization, either via custom or builtin, the bytes ``0x01`` and ``0x02``
            must not appear anywhere. If your payload does contain these bytes or chars, you must
            substitute them prior to this function call.

        Parameters
        ------------
        serializable_object: :class:`object`
            The object to be sent to the receiving end.
        encoding: Optional[:class:`str`]
            The encoding schema of the serialization. If your object serialization results
            in non UTF-8 encoding characters, a different encoding scheme must be used. The
            receiving end must also use this same encoding to decode properly.
            Defaults to ``utf-8``.

        """
        if not self.is_active():
            self._logger.debug(f"Attempted to send data when the link is closed! Ignoring.")
            return

        if type(serializable_object).__name__ not in serialization.IPYC_CUSTOM_SERIALIZATIONS:
            self._logger.debug(f"Serializing {type(serializable_object).__name__} as a default python type")
            serialized_string = str(serializable_object)
        else:
            self._logger.debug(f"Serializing {type(serializable_object).__name__} using a custom defined serialization")
            serialized_string = serialization.IPYC_CUSTOM_SERIALIZATIONS[type(serializable_object).__name__](serializable_object)

        packet = CommunicationPacket(type(serializable_object).__name__, serialized_string)
        self._logger.debug(f"Sending {len(packet.object_serialization)} bytes of '{packet.class_name}'")
        self._connection.send(packet.construct(encoding=encoding))

    def receive(self, encoding='utf-8', return_on_error=False):
        """Receive a serializable object from the other end. If the object is not a custom
        serializable object, python's builtins will be used, otherwise the custom defined
        deserializer will be used.

        Parameters
        ------------
        encoding: Optional[:class:`str`]
            The encoding schema of the serialization. If your object serialization results
            in non UTF-8 encoding characters, a different encoding scheme must be used. The
            receiving end must also use this same encoding to decode properly.
            Defaults to ``utf-8``.
        return_on_error: Optional[:class:`bool`]
            Whether to continue to listen or return if a deserialization error occurred. Otherwise,
            wait until the next valid deserialization occurs.
            Defaults to ``False``.

        Returns
        --------
        Optional[:class:`object`]
            The object that was sent from the sending end. If the deserialization was not successful
            and ``return_on_error`` was set to ``True``, or EOF was encountered resulting in a closed
            connection, ``None`` is returned.
        """
        if not self.is_active():
            self._logger.debug(f"Attempted to read data when link is closed! Returning nothing.")
            return None

        self._logger.debug(f"Waiting for communication from the other side")
        try:
            data = self._connection.recv()
        except (EOFError, ConnectionAbortedError):
            self._logger.debug(f"The downstream connection was aborted")
            self.close()
            return None
        packet = CommunicationPacket.extract(data, encoding=encoding)
        while not packet and not return_on_error:
            self._logger.debug(f"Packet received was not a valid communication packet... waiting for another")
            if self._connection.closed:
                self._logger.debug(f"The downstream connection was closed")
                self.close()
                return None
            try:
                data = self._connection.recv()
            except (EOFError, ConnectionAbortedError):
                self._logger.debug(f"The downstream connection was aborted")
                self.close()
                return None
            packet = CommunicationPacket.extract(data, encoding=encoding)
        if self._connection.closed:
            self._logger.debug(f"The downstream connection as closed")
            self.close()
            return None
        if not packet:
            self._logger.debug(f"Packet received was not a valid communication packet, return_on_error was set to true. Returning.")
            return None

        self._logger.debug(f"Received {len(packet.object_serialization)} bytes of '{packet.class_name}'")
        if packet.class_name not in serialization.IPYC_CUSTOM_DESERIALIZATIONS:
            return eval(packet.class_name)(packet.object_serialization)
        else:
            return serialization.IPYC_CUSTOM_DESERIALIZATIONS[packet.class_name](packet.object_serialization)

    def poll(self, timeout=0.0):
        """Return whether there is any data available to be read from the downstream connection.

        Parameters
        ------------
        timeout: Optional[:class:`float`]
            The number of seconds to wait to see if data is available. If set to ``None``, poll will block
            forever until data is ready to be read. If set to ``0``, return a result immediately.
            Defaults to ``0.0``.

        Returns
        --------
        :class:`bool`
            ``True`` if data is ready to be received, ``False`` otherwise.
        """
        return self._connection.poll(timeout=timeout)


class AsyncIPyCLink:
    """Represents an abstracted async socket connection that handles
    communication between a :class:`AsyncIPyCHost` and a :class:`AsyncIPyCClient`
    This class is internally managed and typically should not be instantiated on
    its own.

    Parameters
    -----------
    reader: :class:`asyncio.StreamReader`
        The managed inbound data reader.
    writer: :class:`asyncio.StreamWriter`
        The managed outbound data writer
    client: Union[:class:`AsyncIPyCHost`, :class:`AsyncIPyCClient`]
        The communication object that is responsible for managing this connection.
    """
    def __init__(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter, client):
        self._reader = reader
        self._writer = writer
        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.debug(f"Established link")
        self._active = True
        self._client = client

    async def close(self):
        """|coro|

        Closes all communication channels with a peer and attempts to send them EOF.
        Informs the parent :class:`AsyncIPyCHost` or :class:`AsyncIPyCClient` of the
        closed connection.
        """
        self._logger.debug(f"Beginning to close link")
        self._reader = None
        if self._writer.can_write_eof():
            self._writer.write_eof()
            try:
                await self._writer.drain()
            except ConnectionAbortedError:
                pass
        self._writer.close()
        if sys.version_info >= (3, 7):
            await self._writer.wait_closed()
        self._writer = None
        self._active = False
        self._client.connections.remove(self)
        self._logger.debug(f"Closed link")

    def is_active(self):
        """:class:`bool`: Indicates if the communication channels are closed, at EOF, or no longer viable."""
        # Quickly check if the state of the reader changed from the remote
        if not self._reader or self._reader.at_eof() or not self._writer:
            self._active = False
        return self._active

    async def send(self, serializable_object: object, drain_immediately=True, encoding='utf-8'):
        """|coro|

        Send a serializable object to the receiving end. If the object is not a custom
        serializable object, python's builtins will be used. If the object is a custom
        serializable, the receiving end must also have this object in their list of custom
        deserializers.

        .. warning::
            After the result of serialization, either via custom or builtin, the bytes ``0x01`` and ``0x02``
            must not appear anywhere. If your payload does contain these bytes or chars, you must
            substitute them prior to this function call.

        Parameters
        ------------
        serializable_object: :class:`object`
            The object to be sent to the receiving end.
        drain_immediately: Optional[:class:`bool`]
            Whether to flush the output buffer right now or not.
            Defaults to ``True``.
        encoding: Optional[:class:`str`]
            The encoding schema of the serialization. If your object serialization results
            in non UTF-8 encoding characters, a different encoding scheme must be used. The
            receiving end must also use this same encoding to decode properly.
            Defaults to ``utf-8``.

        """
        if not self.is_active():
            self._logger.debug(f"Attempted to send data when the writer or link is closed! Ignoring.")
            return

        if type(serializable_object).__name__ not in serialization.IPYC_CUSTOM_SERIALIZATIONS:
            self._logger.debug(f"Serializing {type(serializable_object).__name__} as a default python type")
            serialized_string = str(serializable_object)
        else:
            self._logger.debug(f"Serializing {type(serializable_object).__name__} using a custom defined serialization")
            serialized_string = serialization.IPYC_CUSTOM_SERIALIZATIONS[type(serializable_object).__name__](serializable_object)

        packet = CommunicationPacket(type(serializable_object).__name__, serialized_string)
        self._logger.debug(f"Sending {len(packet.object_serialization)} bytes of '{packet.class_name}'")
        self._writer.write(packet.construct(encoding=encoding))
        if drain_immediately:
            self._logger.debug(f"Draining the writer")
            await self._writer.drain()

    async def receive(self, encoding='utf-8', return_on_error=False):
        """|coro|

        Receive a serializable object from the other end. If the object is not a custom
        serializable object, python's builtins will be used, otherwise the custom defined
        deserializer will be used.

        Parameters
        ------------
        encoding: Optional[:class:`str`]
            The encoding schema of the serialization. If your object serialization results
            in non UTF-8 encoding characters, a different encoding scheme must be used. The
            receiving end must also use this same encoding to decode properly.
            Defaults to ``utf-8``.
        return_on_error: Optional[:class:`bool`]
            Whether to continue to listen or return if a deserialization error occurred. Otherwise,
            wait until the next valid deserialization occurs.
            Defaults to ``False``.

        Returns
        --------
        Optional[:class:`object`]
            The object that was sent from the sending end. If the deserialization was not successful
            and ``return_on_error`` was set to ``True``, or EOF was encountered resulting in a closed
            connection, ``None`` is returned.
        """
        if not self.is_active():
            self._logger.debug(f"Attempted to read data when the writer or link is closed! Returning nothing.")
            return None

        self._logger.debug(f"Waiting for communication from the other side")
        try:
            data = await self._reader.readline()
        except ConnectionAbortedError:
            self._logger.debug(f"The downstream connection was aborted")
            await self.close()
            return None
        packet = CommunicationPacket.extract(data, encoding=encoding)
        while not packet and not return_on_error:
            self._logger.debug(f"Packet received was not a valid communication packet... waiting for another")
            if self._reader.at_eof():
                self._logger.debug(f"The downstream writer closed the connection")
                await self.close()
                return None
            try:
                data = await self._reader.readline()
            except ConnectionAbortedError:
                self._logger.debug(f"The downstream connection was aborted")
                await self.close()
                return None
            packet = CommunicationPacket.extract(data, encoding=encoding)
        if self._reader.at_eof():
            self._logger.debug(f"The downstream writer closed the connection")
            await self.close()
            return None
        if not packet:
            self._logger.debug(f"Packet received was not a valid communication packet, return_on_error was set to true. Returning.")
            return None

        self._logger.debug(f"Received {len(packet.object_serialization)} bytes of '{packet.class_name}'")
        if packet.class_name not in serialization.IPYC_CUSTOM_DESERIALIZATIONS:
            return eval(packet.class_name)(packet.object_serialization)
        else:
            return serialization.IPYC_CUSTOM_DESERIALIZATIONS[packet.class_name](packet.object_serialization)



import logging

from multiprocessing.connection import Listener, Client

from .links import IPyCLink


class IPyCHost:
    """Represents an abstracted synchronous socket listener that connects with
    and listens to :class:`IPyCClient` clients.
    A number of options can be passed to the :class:`IPyCHost`.

    Parameters
    -----------
    ip_address: Optional[:class:`str`]
        The IP address start listening on. This defaults to ``localhost``.
    port: Optional[:class:`int`]
        The port the listener binds to. This defaults to ``9999``. Check
        your system to make sure this port is not used by another service.
        To use multiple :class:`IPyCHost` hosts, ensure the ports are
        different between instantiations.
    """
    def __init__(self, ip_address: str='localhost', port:  int=9999):
        self._ip_address = ip_address
        self._port = port
        self._logger = logging.getLogger(self.__class__.__name__)
        self._logger.info(f"Binding to address {ip_address}:{port}")
        self._server = Listener((ip_address, port))
        self._closed = False
        self._connections = set()

    def is_closed(self):
        """:class:`bool`: Indicates if the underlying socket listener is closed or no longer listening."""
        return self._closed

    @property
    def connections(self):
        """:class:`set`: Returns the set of all active :class:`IPyCLink` connections the host is handling."""
        return self._connections

    def close(self):
        """Closes all :class:`IPyCLink` connections and stops the internal listener."""
        if self._closed:
            return
        self._closed = True
        for connection in list(self.connections):
            connection.close()
        self._server.close()

    def wait_for_client(self) -> IPyCLink:
        """Starts listening for :class:`IPyCClient` clients to connect.

        Returns
        -------
        :class:`IPyCLink`
            The connection that has been established with a :class:`IPyCClient`.
        """
        self._logger.info("Starting to wait for a client...")
        if not self.is_closed():
            connection = IPyCLink(self._server.accept(), self)
            self._connections.add(connection)
            return connection


class IPyCMaster(IPyCHost):
    """Pseudo-class for IPyCHost"""
    pass


class IPyCClient:
    """Represents an abstracted synchronous socket client that connects to
    and communicates with :class:`IPyCHost` hosts.
    A number of options can be passed to the :class:`IPyCClient`.

    Parameters
    -----------
    ip_address: Optional[:class:`str`]
        The IP address to connect to. This defaults to ``localhost``.
    port: Optional[:class:`int`]
        The port to target at the host IP address. This defaults to ``9999``.
    """
    def __init__(self, ip_address: str='localhost', port:  int=9999):
        self._ip_address = ip_address
        self._port = port
        self._logger = logging.getLogger(self.__class__.__name__)
        self._link = None
        self._closed = False

    def close(self):
        """Closes the :class:`IPyCLink` connection and stops the client connection."""
        if self._closed:
            return
        self._closed = True
        if self._link:
            self._link.close()
        self._link = None

    def connect(self) -> IPyCLink:
        """A shorthand coroutine for `asyncio.open_connection`. Any
        arguments supplied are directly fed to that method.
        See the asyncio documentation for these arguments and their use.

        Returns
        -------
        :class:`IPyCLink`
            The connection that has been established with a :class:`IPyCHost`.
        """
        self._logger.info("Starting to connect to the host...")
        connection = Client((self._ip_address, self._port))
        self._link = IPyCLink(connection, self)
        return self._link

    @property
    def connections(self):
        """:class:`set`: Returns the set of all active :class:`IPyCLink` connections the client is handling.
        The number of connections is always either one or none.
        """
        return {self._link} if self._link else None


class IPyCSlave(IPyCClient):
    """Pseudo-class for IPyCClient"""
    pass



import json

IPYC_CUSTOM_SERIALIZATIONS = {
    dict.__name__: json.dumps,
}

IPYC_CUSTOM_DESERIALIZATIONS = {
    dict.__name__: json.loads,
}


def add_custom_serialization(class_object: object, class_serializer):
    """Register a serialization function for a particular object class. Only
    one serialization method per object is allowed. The method must be a non-
    blocking function.

    .. warning::
        If a custom serializer is defined, ensure a deserializer is defined as well on the receiving end.
        These methods could be ``@classmethod`` methods defined in the class for which you wish to transfer.

    Parameters
    ------------
    class_object: :class:`object`
        The class or object type that will be serialized.
    class_serializer: function
        The function to call when serialization is requested for the object. Must return a string.

    """
    IPYC_CUSTOM_SERIALIZATIONS[class_object.__name__] = class_serializer


def update_custom_serialization(class_object: object, class_serializer):
    """Wrapped method for :meth:`add_custom_serialization`"""
    add_custom_serialization(class_object, class_serializer)


def remove_custom_serialization(class_object: object):
    """De-register a serialization function for a particular object class.

    .. warning::
        If a custom serializer is removed, serialization of this object will then
        default to python's builtins.

    Parameters
    ------------
    class_object: :class:`object`
        The class or object type that will be removed from the custom serializer dictionary.

    """
    if class_object.__name__ in IPYC_CUSTOM_SERIALIZATIONS:
        del IPYC_CUSTOM_SERIALIZATIONS[class_object.__name__]


def add_custom_deserialization(class_object: object, class_deserializer):
    """Register a deserialization function for a particular object class. Only
    one deserialization method per object is allowed. The method must be a non-
    blocking function.

    .. warning::
        If a custom deserializer is defined, ensure a serializer is defined as well on the sending end.
        These methods could be ``@classmethod`` methods defined in the class for which you wish to transfer.

    Parameters
    ------------
    class_object: :class:`object`
        The class or object type that will be serialized.
    class_deserializer: function
        The function to call when deserialization is requested for the object. Must return a :class:`object`.

    """
    IPYC_CUSTOM_DESERIALIZATIONS[class_object.__name__] = class_deserializer


def update_custom_deserialization(class_object: object, class_deserializer):
    """Wrapped method for :meth:`add_custom_deserialization`"""
    add_custom_deserialization(class_object, class_deserializer)


def remove_custom_deserialization(class_object: object):
    """De-register a deserialization function for a particular object class.

    .. warning::
        If a custom deserializer is removed, deserialization of this object will then
        default to python's builtins.

    Parameters
    ------------
    class_object: :class:`object`
        The class or object type that will be removed from the custom deserializer dictionary.

    """
    if class_object.__name__ in IPYC_CUSTOM_DESERIALIZATIONS:
        del IPYC_CUSTOM_DESERIALIZATIONS[class_object.__name__]



# -*- coding: utf-8 -*-

"""
Python IPC (Inter-Process Communication) Wrapper and Implementation
~~~~~~~~~~~~~~~~~~~

A basic IPC implementation for Python 3

:copyright: (c) 2020-2020 dovedevic
:license: MIT, see LICENSE for more details.

"""

__title__ = 'ipyc'
__author__ = 'dovedevic'
__license__ = 'MIT'
__copyright__ = 'Copyright 2020-present dovedevic'
__version__ = '1.1.1'

__path__ = __import__('pkgutil').extend_path(__path__, __name__)

from collections import namedtuple
import logging

from .blocking import IPyCHost, IPyCMaster, IPyCClient, IPyCSlave
from .asynchronous import AsyncIPyCHost, AsyncIPyCMaster, AsyncIPyCClient, AsyncIPyCSlave
from .links import IPyCLink, AsyncIPyCLink
from . import serialization as IPyCSerialization

VersionInfo = namedtuple('VersionInfo', 'major minor micro releaselevel serial')
version_info = VersionInfo(major=1, minor=1, micro=1, releaselevel='release', serial=0)

logging.getLogger(__name__).addHandler(logging.NullHandler())



class CommunicationPacket:
    def __init__(self, class_name: str, object_serialization: str):
        self.__name = class_name
        self.__object_serialization = object_serialization

    @property
    def class_name(self):
        return self.__name

    @property
    def object_serialization(self):
        return self.__object_serialization

    def construct(self, encoding: str) -> bytes:
        return "\x01{}\x02{}\n".format(self.__name, self.__object_serialization.replace('\n', '\x1a')).encode(encoding)

    @staticmethod
    def extract(packet: bytes, encoding: str):
        try:
            packet = packet.decode(encoding)
            if not packet.startswith('\x01') or not packet.endswith('\n'):
                raise ValueError
            serialization_string = packet[1:-1]
            name, serialization = serialization_string.split('\x02')
            return CommunicationPacket(name, serialization.replace('\x1a', '\n'))
        except (MemoryError, RuntimeError, ValueError, UnicodeError, UnicodeDecodeError):
            return None



from setuptools import setup
import re

requirements = []
with open('requirements.txt') as f:
  requirements = f.read().splitlines()

with open('ipyc/__init__.py') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', f.read(), re.MULTILINE).group(1)

with open('README.rst') as f:
    readme = f.read()


extras_require = {
    'docs': [
        'sphinx==3.0.3',
        'sphinxcontrib_trio==1.1.2',
        'sphinxcontrib-websupport',
    ]
}


setup(name='IPyC',
      author='dovedevic',
      url='https://github.com/dovedevic/IPyC',
      project_urls={
        "Documentation": "https://ipyc.readthedocs.io/",
        "Issue tracker": "https://github.com/dovedevic/IPyC/issues",
      },
      version=version,
      packages=['ipyc'],
      license='MIT',
      description='A basic IPC implementation for Python',
      long_description=readme,
      long_description_content_type="text/x-rst",
      install_requires=requirements,
      python_requires='>=3.5.3',
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
      ]
)
